// ***************************************************************************
// 			AFLOW Compare Structure - Functions
//		David Hicks (d.hicks@duke.edu) and Carlo de Santo
// ***************************************************************************
#include <fstream>
#include <iostream>
#include <vector>
#include <string>
#include "aflow.h"
#include "aflow_pflow.h"
#include "aflow_compare_structure.h"
//[THREADS]#include <thread>
//[THREADS]#include <atomic>
//#include <stdatomic.h> // Needed to communicate between threads

// ***************************************************************************
// Prototype Class
// ***************************************************************************
// ===== Constructor ===== //
StructurePrototype::StructurePrototype() {
  iomode = JSON_MODE;
  master_structure_name = "";
  master_structure.Clear();
  elements.clear();
  stoichiometry.clear();
  pearson = "";
  space_group = 0;
  proto_structures_names.clear();
  proto_structures.clear();
  misfits.clear();
}

// ===== Free  ===== //
void StructurePrototype::Free() {
}

// ===== Destructor ===== //
StructurePrototype::~StructurePrototype() {
  master_structure.Clear();
  elements.clear();
  stoichiometry.clear();
  proto_structures_names.clear();
  proto_structures.clear();
  misfits.clear();
  Free();
}

// ===== Copy Constructor ===== //
StructurePrototype::StructurePrototype(const StructurePrototype& b) {
  Copy(b);
}

// ===== Copy Constructor Function ===== //
void StructurePrototype::Copy(const StructurePrototype& b) {
  if(this != &b) {
    iomode = b.iomode;
    master_structure_name = b.master_structure_name;
    master_structure = b.master_structure;
    elements = b.elements;
    stoichiometry = b.stoichiometry;
    pearson = b.pearson;
    space_group = b.space_group;
    proto_structures_names = b.proto_structures_names;
    proto_structures = b.proto_structures;
    misfits = b.misfits;
  }
}

// ===== Assignment Operator (operator=) ===== //
const StructurePrototype& StructurePrototype::operator=(const StructurePrototype& b) {
  if(this != &b) {
    Free();
    Copy(b);
  }
  return *this;
}

// ===== Output Handler ===== //
ostream& operator<<(ostream& oss, const StructurePrototype& StructurePrototype) {
  //if(StructurePrototype.iomode!=JSON_MODE){ //A safeguard until we construct more output schemes.
  //  StructurePrototype.iomode=JSON_MODE;
  //}
  if(StructurePrototype.iomode == JSON_MODE) {
    oss << "{" << endl;
    oss << "  \"master_structure\": \"" << StructurePrototype.master_structure_name << "\"," << endl;
    oss << "  \"elements\": [";
    for (uint i = 0; i < StructurePrototype.elements.size(); i++) {
      if(i == 0) {
        oss << endl
            << "    \"" << StructurePrototype.elements[i] << "\"";
      } else {
        oss << "," << endl
            << "    \"" << StructurePrototype.elements[i] << "\"";
      }
    }
    oss << endl
        << "  ]," << endl;
    oss << "  \"stoichiometry\": \"";
    for (uint i = 0; i < StructurePrototype.stoichiometry.size(); i++) {
      if(i == 0) {
        oss << StructurePrototype.stoichiometry[i];
      } else {
        oss << ":" << StructurePrototype.stoichiometry[i];
      }
    }
    oss << "\"," << endl;
    oss << "  \"pearson\": \"" << StructurePrototype.pearson << "\"," << endl;
    oss << "  \"space_group\": \"" << StructurePrototype.space_group << "\"," << endl;
    oss << "  \"proto_structures\": [";  // << endl;
    for (uint i = 0; i < StructurePrototype.proto_structures_names.size(); i++) {
      if(i == 0) {
        oss << endl
            << "    \"" << StructurePrototype.proto_structures_names[i] << "\"";
      } else {
        oss << "," << endl
            << "    \"" << StructurePrototype.proto_structures_names[i] << "\"";
      }
    }
    if(StructurePrototype.proto_structures_names.size() != 0) {
      oss << endl
          << "  ]," << endl;
    } else {
      oss << "]," << endl;
    }
    oss << "  \"misfits\": [";  // << endl;
    for (uint i = 0; i < StructurePrototype.misfits.size(); i++) {
      if(i == 0) {
        oss << endl
            << "    \"" << StructurePrototype.misfits[i] << "\"";
      } else {
        oss << "," << endl
            << "    \"" << StructurePrototype.misfits[i] << "\"";
      }
    }
    if(StructurePrototype.misfits.size() != 0) {
      oss << endl
          << "  ]" << endl;
    } else {
      oss << "]" << endl;
    }
    oss << "}";
  }
  return oss;
}

// ***************************************************************************
// END:: Prototype Class
// ***************************************************************************

// ***************************************************************************

// ***************************************************************************
// *                                                                         *
// *                             FUNCTIONS                                   *
// *                                                                         *
// ***************************************************************************

// ***************************************************************************
//  Get Stoichiometry - Obtain stoichiometries from xstructure
// ***************************************************************************
namespace compare {
vector<int> getStoichiometry(const xstructure& xstr, const bool& same_species) {
  // Obtains the least common multiple representation of the stoichiometry.

  vector<int> stoich;
  if(xstr.species.size() == 1) {
    stoich.push_back(1);
  } else {
    stoich = gcdStoich(xstr.num_each_type);
  }
  // If a structure prototype comparison (not material type), ensure
  // stoichiometries are in numerical order for comparison.  Else,
  // leave in position indicating atomic species count.
  if(same_species == false) {
    for (uint i = 0; i < stoich.size(); i++) {
      std::sort(stoich.begin(), stoich.end());
    }
  }
  return stoich;
}
}

// ***************************************************************************
//  Get Elements
// ***************************************************************************
namespace compare {
vector<string> getElements(xstructure& xstr) {
  // Obtains the elements in the structure.

  bool LDEBUG = (false || XHOST.DEBUG);
  vector<string> velements;
  // If atoms in poscar not labeled in either POSCAR; assign fake names
  if(xstr.atoms.at(0).name == "") {
    if(LDEBUG) { cerr << "compare:: "
                       << "WARNING!!!!! Atoms not labeled ... Assigning Fake names" << endl; }
    fakeAtomsName(xstr);
  }
  string prev_element = "";
  for (uint i = 0; i < xstr.atoms.size(); i++) {
    if(xstr.atoms[i].name != prev_element) {
      velements.push_back(xstr.atoms[i].name);
      prev_element = xstr.atoms[i].name;
    }
  }
  return velements;
}
}

// ***************************************************************************
// gcdStoich - Euler's Greatest Common Divisor Algorithm
// ***************************************************************************
namespace compare {
vector<int> gcdStoich(const deque<int>& numbers) {
  // This is Euler's Greates Common Divisor Algorithm.  It is used to determine
  // the least common multiple representation for the stoichiometry.

  int global_GCD;
  int GCD;
  vector<int> reduced_numbers;
  // Find min number first
  int min = 0;
  for (uint i = 0; i < numbers.size(); i++) {
    if(i == 0) {
      min = numbers[i];
    } else {
      if(numbers[i] < min) {
        min = numbers[i];
      }
    }
  }
  bool found_GCD = true;
  for (uint i = 0; i < numbers.size(); i++) {
    if(numbers[i] % min != 0) {
      found_GCD = false;
      break;
    }
  }
  if(found_GCD == true) {
    global_GCD = min;
  } else if(found_GCD == false) {
    int remainder = 1000;
    int divisor = min;
    for (uint i = 0; i < numbers.size(); i++) {
      int num = numbers[i];
      while (remainder != 0) {
        remainder = (num % divisor);
        if(remainder == 0) {
          GCD = divisor;
        } else {
          num = divisor;
          divisor = remainder;
        }
      }
      divisor = GCD;
      if(i == 0) {
        global_GCD = GCD;
      } else if(GCD < global_GCD) {
        global_GCD = GCD;
      }
      remainder = 100;
    }
  }
  for (uint i = 0; i < numbers.size(); i++) {
    reduced_numbers.push_back(numbers[i] / global_GCD);
    if(numbers[i] % global_GCD) {
      cerr << "compare::ERROR - Error in GCD procedure. Exiting. (David Hicks: d.hicks@duke.edu)" << endl;
      exit(1);
    }
  }
  return reduced_numbers;
}
}

// ***************************************************************************
// calculateSymmetry - Calculate Symmetry
// ***************************************************************************
namespace compare {
void calculateSymmetry(xstructure& xstr, vector<string>& vpearsons, vector<uint>& vsgroups) {
  // Calculates the Pearson symbol and space group of the structure

  xstr.GetLatticeType();  //A little slow. Consider a different method -> (SpaceGroup_ITC) finds this
  vpearsons.push_back(xstr.pearson_symbol);
  vsgroups.push_back(xstr.SpaceGroup_ITC());
}
}

// ***************************************************************************
// createStructurePrototypes - Group structures by Pearson symbol, then space group
// ***************************************************************************
namespace compare {
void createStructurePrototypes(vector<StructurePrototype>& comparison_schemes,
                               const vector<xstructure>& vxstrs, const bool& same_species,
                               const vector<vector<string> >& vvelements,
                               vector<vector<int> >& vstoichs, vector<string>& vpearsons,
                               vector<uint>& vsgroups, const string& directory,
                               const vector<string>& vfiles) {
  // Populates the structure information into the StructurePrototype object.
  // It groups structure based on their stoichiometry and pearson symbol and
  // space group. A "master" structure is chosen and will be compared to the
  // possible "duplicates". The misfit values are set to -1.0 until compared.

  // First, separate by stoichiometry
  for (uint i = 0; i < vstoichs.size(); i++) {
    bool scheme_created = false;
    if(i == 0) {
      StructurePrototype tmp;
      tmp.master_structure_name = directory + "/" + vfiles[i];
      tmp.elements = vvelements[i];
      tmp.stoichiometry = vstoichs[i];
      tmp.pearson = vpearsons[i];
      tmp.space_group = vsgroups[i];
      tmp.master_structure = vxstrs[i];
      comparison_schemes.push_back(tmp);
    } else {
      for (uint j = 0; j < comparison_schemes.size(); j++) {
        bool same_material_stoich = false;
        ostringstream tmp;
        tmp.clear();
        if(same_species == true &&
            matchableSpecies(vxstrs[i], comparison_schemes[j].master_structure, same_species) == true) {
          same_material_stoich = true;
        } else if(same_species == false) {
          same_material_stoich = true;
        }
        if(same_material_stoich == true &&
            vstoichs[i] == comparison_schemes[j].stoichiometry &&
            vpearsons[i] == comparison_schemes[j].pearson &&
            vsgroups[i] == comparison_schemes[j].space_group) {
          if(same_species == false) {
            for (uint e = 0; e < vvelements[i].size(); e++) {
              bool already_in = false;
              for (uint f = 0; f < comparison_schemes[j].elements.size(); f++) {
                if(vvelements[i][e] == comparison_schemes[j].elements[f]) {
                  already_in = true;
                  break;
                }
              }
              if(already_in == false) {
                comparison_schemes[j].elements.push_back(vvelements[i][e]);
              }
            }
          }
          comparison_schemes[j].proto_structures_names.push_back(directory + "/" + vfiles[i]);
          comparison_schemes[j].proto_structures.push_back(vxstrs[i]);
          comparison_schemes[j].misfits.push_back(-1.0);
          scheme_created = true;
          break;
        }
      }
      if(scheme_created == false) {
        StructurePrototype tmp;
        tmp.master_structure_name = directory + "/" + vfiles[i];
        tmp.elements = vvelements[i];
        tmp.stoichiometry = vstoichs[i];
        tmp.pearson = vpearsons[i];
        tmp.space_group = vsgroups[i];
        tmp.master_structure = vxstrs[i];
        comparison_schemes.push_back(tmp);
      }
    }
  }
}
}

// ***************************************************************************
// NumberMismaches - Count the number of non-matches
// ***************************************************************************
namespace compare {
int numberMismatches(const vector<StructurePrototype> comparison_schemes) {
  int num_mismatches = 0;
  for (uint i = 0; i < comparison_schemes.size(); i++) {
    for (uint j = 0; j < comparison_schemes[i].misfits.size(); j++) {
      if(comparison_schemes[i].misfits[j] > 0.1 || comparison_schemes[i].misfits[j] == -1.0) {
        num_mismatches += 1;
      }
    }
  }
  return num_mismatches;
}
}

// ***************************************************************************
// appendStructurePrototypes - Create new structure prototypes after comparisons
// ***************************************************************************
namespace compare {
void appendStructurePrototypes(vector<StructurePrototype>& comparison_schemes,
                               vector<StructurePrototype>& final_prototypes) {
  // This cleans the StrucuturePrototype objects be removing all the mismatches.
  // Then, it takes the mismatches and makes them into new StructurePrototype objects
  // to be compared.

  vector<StructurePrototype> tmp_list;
  for (uint i = 0; i < comparison_schemes.size(); i++) {
    bool first_mismatch = true;
    for (uint j = 0; j < comparison_schemes[i].misfits.size(); j++) {
      if(comparison_schemes[i].misfits[j] > 0.1 || comparison_schemes[i].misfits[j] == -1.0) {
        // Take first mismatch and make as the master structure in the new object
        if(first_mismatch == true) {
          StructurePrototype tmp;
          tmp.master_structure_name = comparison_schemes[i].proto_structures_names[j];
          tmp.stoichiometry = comparison_schemes[i].stoichiometry;
          tmp.pearson = comparison_schemes[i].pearson;
          tmp.space_group = comparison_schemes[i].space_group;
          tmp.master_structure = comparison_schemes[i].proto_structures[j];
          tmp_list.push_back(tmp);
          comparison_schemes[i].proto_structures_names.erase(comparison_schemes[i].proto_structures_names.begin() + j);
          comparison_schemes[i].proto_structures.erase(comparison_schemes[i].proto_structures.begin() + j);
          comparison_schemes[i].misfits.erase(comparison_schemes[i].misfits.begin() + j);
          j--;
          first_mismatch = false;
        }
        // If not the first mismatch, add as a proto structure in the new object
        else if(first_mismatch == false) {
          tmp_list.back().proto_structures_names.push_back(comparison_schemes[i].proto_structures_names[j]);
          tmp_list.back().proto_structures.push_back(comparison_schemes[i].proto_structures[j]);
          tmp_list.back().misfits.push_back(-1.0);
          comparison_schemes[i].proto_structures_names.erase(comparison_schemes[i].proto_structures_names.begin() + j);  //
          comparison_schemes[i].proto_structures.erase(comparison_schemes[i].proto_structures.begin() + j);              //
          comparison_schemes[i].misfits.erase(comparison_schemes[i].misfits.begin() + j);                                //
          j--;
        }
      }
    }
    // Store finished (already compared) schemes in final_prototypes
    final_prototypes.push_back(comparison_schemes[i]);
    comparison_schemes.erase(comparison_schemes.begin() + i);
    i--;
  }
  // Store newly generated schemes (not compared yet) into comparison_schemes
  comparison_schemes = tmp_list;
}
}

// ***************************************************************************
// checkPrototypes - Ensure prototypes of different SG are compared
// ***************************************************************************
namespace compare {
void checkPrototypes(const int& num_proc, const bool& same_species,
                     vector<StructurePrototype>& final_prototypes) {
  // Checks to see if prototypes of different space groups are similar.
  // If they are, then we combine the StructurePrototype objects into one.
  // When combining, we keep the "master" prototype as the one with a higher
  // symmetry (i.e. higher space group).

  bool scale_volume=true; //default is true
  for (uint i = 0; i < final_prototypes.size(); i++) {
    vector<int> store_indices;
    vector<double> store_misfits;
    int min_index = -1;
    double min_misfit = -1.0;
    for (uint j = i; j < final_prototypes.size(); j++) {
      ostringstream tmp_oss;
      tmp_oss.clear();
      if(
          // If same_species==true
          (same_species == true &&
           matchableSpecies(final_prototypes[i].master_structure, final_prototypes[j].master_structure, same_species) == true &&
           final_prototypes[i].stoichiometry == final_prototypes[j].stoichiometry &&
           final_prototypes[i].pearson == final_prototypes[j].pearson &&
           final_prototypes[i].space_group != final_prototypes[j].space_group) ||
          // If same_species==false
          (same_species == false &&
           final_prototypes[i].stoichiometry == final_prototypes[j].stoichiometry &&
           final_prototypes[i].pearson == final_prototypes[j].pearson &&
           final_prototypes[i].space_group != final_prototypes[j].space_group)) {
        double final_misfit = -1.0;
        aflowCompareStructure(num_proc, final_prototypes[i].master_structure,
                              final_prototypes[j].master_structure, same_species, scale_volume,
                              tmp_oss, final_misfit);
        if(final_misfit < min_misfit) {
          min_misfit = final_misfit;
          min_index = j;
        }
      }
    }
    // If one prototype is similar to another, add to one with higher space group
    if(min_misfit != -1.0) {
      int sg_ind = -1;
      uint other_ind = -1;
      if(final_prototypes[i].space_group > final_prototypes[min_index].space_group) {
        sg_ind = i;
        other_ind = min_index;
      } else {
        sg_ind = min_index;
        other_ind = i;
      }
      // Transfer info to prototype with higher space group
      final_prototypes[sg_ind].proto_structures_names.push_back(final_prototypes[other_ind].master_structure_name);
      final_prototypes[sg_ind].proto_structures_names.insert(final_prototypes[sg_ind].proto_structures_names.end(),
                                                             final_prototypes[other_ind].proto_structures_names.begin(),
                                                             final_prototypes[other_ind].proto_structures_names.end());
      final_prototypes[sg_ind].proto_structures.push_back(final_prototypes[other_ind].master_structure);
      final_prototypes[sg_ind].proto_structures.insert(final_prototypes[sg_ind].proto_structures.end(),
                                                       final_prototypes[other_ind].proto_structures.begin(),
                                                       final_prototypes[other_ind].proto_structures.end());
      // Delete the prototype with the lower space group
      final_prototypes.erase(final_prototypes.begin() + other_ind);
      // If the index deleted was less than the initial loop (i), then need to reduce iterator
      if(other_ind <= i) {
        i--;
      }
    }
  }
}
}

// ***************************************************************************
// prepareTextOutput - Displays results for .txt file
// ***************************************************************************
namespace compare {
void prepareTextOutput(ostream& ss_out, const bool& same_species,
                       const vector<StructurePrototype>& final_prototypes) {
  // Displays comparison information in an TXT file.

  for (uint j = 0; j < final_prototypes.size(); j++) {
    ss_out << "==============================================================" << endl;
    ss_out << "# ";
    if(same_species == true) {
      for (uint k = 0; k < final_prototypes[j].elements.size(); k++) {
        ss_out << final_prototypes[j].elements[k] << final_prototypes[j].stoichiometry[k];
      }
      ss_out << "   " << final_prototypes[j].pearson
             << "   " << final_prototypes[j].space_group
             << "   "
             << "(" << final_prototypes[j].master_structure_name << ")" << endl;
    } else if(same_species == false) {
      for (uint k = 0; k < final_prototypes[j].stoichiometry.size(); k++) {
        if(k == 0) {
          ss_out << final_prototypes[j].stoichiometry[k];
        } else {
          ss_out << ":" << final_prototypes[j].stoichiometry[k];
        }
      }
      ss_out << "   " << final_prototypes[j].pearson
             << "   " << final_prototypes[j].space_group
             << "   "
             << "(" << final_prototypes[j].master_structure_name << ")" << endl;
    }
    ss_out << "==============================================================" << endl;
    for (uint k = 0; k < final_prototypes[j].proto_structures_names.size(); k++) {
      ss_out << "   " << setw(80) << std::left << final_prototypes[j].proto_structures_names[k]
             << setw(15) << std::left << final_prototypes[j].misfits[k] << endl;
    }
  }
}
}

// ***************************************************************************
// Matchable species
// ***************************************************************************
namespace compare {
bool matchableSpecies(const xstructure& xstr1, const xstructure& xstr2,
                      const bool& same_species) {
  // Determine if it is possible to match up species based on the number of atoms
  // (i.e reduced stoichiometries equal)

  bool LDEBUG = (false || XHOST.DEBUG);
  vector<int> stoich1;
  vector<int> stoich2;
  bool matchable = true;
  if(xstr1.species.size() == xstr2.species.size()) {
    if(xstr1.species.size() == 1) {
      stoich1.push_back(1);
      stoich2.push_back(1);
    } else {
      stoich1 = gcdStoich(xstr1.num_each_type);
      stoich2 = gcdStoich(xstr2.num_each_type);
    }
    // Check if we can match to same species (atoms and stoichs)
    if(same_species == true) {
      for (uint i = 0; i < stoich1.size(); i++) {
        if(stoich1[i] != stoich2[i] || xstr1.species[i] != xstr2.species[i]) {
          matchable = false;
          break;
        }
      }
    }
    // Check if we can match stoichs only
    else {
      for (uint i = 0; i < stoich1.size(); i++) {
        std::sort(stoich1.begin(), stoich1.end());
        std::sort(stoich2.begin(), stoich2.end());
      }
      for (uint i = 0; i < stoich1.size(); i++) {
        if(stoich1[i] != stoich2[i]) {
          matchable = false;
          break;
        }
      }
    }
    if(matchable == true) {
      return true;
    } else {
      return false;
    }
  } else {
    if(LDEBUG) { cerr << "compare:: "
                       << "NUMBER OF TYPES OF ATOMIC SPECIES:   xstr1:  " << xstr1.num_each_type.size()
                       << "           xstr2:  " << xstr2.num_each_type.size() << endl; }
    if(LDEBUG) { cerr << "compare:: "
                       << "NUMBER OF TYPES OF ATOMIC SPECIES IS NOT THE SAME...QUITTING..." << endl; }
    return false;
  }
}
}

// ***************************************************************************
// Same Species
// ***************************************************************************
namespace compare {
bool sameSpecies(const xstructure& xstr1, const xstructure& xstr2, const bool& display) {
  // Determines if the same atomic species are present in both structures with
  // the same multiplicity. Useful for initial screening of duplicates.

  bool LDEBUG = (false || XHOST.DEBUG);
  uint matches = 0;
  vector<int> num_atoms1;
  vector<int> name_indexstr1;
  vector<int> num_atoms2;
  vector<int> name_indexstr2;
  bool duplicatefound = false;

  // Check for same number of species in x
  if(xstr1.num_each_type.size() == xstr2.num_each_type.size()) {
    num_atoms1.push_back(1);
    name_indexstr1.push_back(0);

    //Organize into vectors with names and multi: xstr1
    for (uint i = 1; i < xstr1.atoms.size(); i++) {
      for (uint j = 0; j < num_atoms1.size(); j++) {
        //Check if the names are the same
        if(xstr1.atoms.at(i).name == xstr1.atoms.at(name_indexstr1.at(j)).name) {
          //If so, add to multiplicity of atom type
          num_atoms1.at(j) += 1;
          duplicatefound = true;
        }
      }
      if(duplicatefound == false) {
        //If not, make new entry new atom
        num_atoms1.push_back(1);
        name_indexstr1.push_back(i);
      }
      duplicatefound = false;
    }
    num_atoms2.push_back(1);
    name_indexstr2.push_back(0);

    //Same procedure as above for xstr2
    for (uint i = 1; i < xstr2.atoms.size(); i++) {
      for (uint j = 0; j < num_atoms2.size(); j++) {
        if(xstr2.atoms.at(i).name == xstr2.atoms.at(name_indexstr2.at(j)).name) {
          num_atoms2.at(j) += 1;
          duplicatefound = true;
        }
      }
      if(duplicatefound == false) {
        num_atoms2.push_back(1);
        name_indexstr2.push_back(i);
      }
      duplicatefound = false;
    }

    //Check to see if name and multiplicity is the same for xstr1 and xstr2 for each vector entry
    if(num_atoms1.size() == num_atoms2.size()) {
      for (uint i = 0; i < num_atoms1.size(); i++) {
        for (uint j = 0; j < num_atoms2.size(); j++) {
          if(num_atoms1.at(i) == num_atoms2.at(j) &&
              xstr1.atoms.at(name_indexstr1.at(i)).name == xstr2.atoms.at(name_indexstr2.at(j)).name) {
            matches++;
          }
        }
      }
    }
    // Display counts
    if(display == true) {
      if(LDEBUG) {
        for (uint i = 0; i < num_atoms1.size(); i++) {
          cerr << "compare::xstr1: Atom type- " << xstr1.atoms.at(name_indexstr1.at(i)).name << " (" << num_atoms1.at(i) << ")  "
               << "xstr2: Atom type- " << xstr2.atoms.at(name_indexstr2.at(i)).name << " (" << num_atoms2.at(i) << ")" << endl;
        }
      }
    }
  } else {
    // Display counts
    if(display == true) {
      if(LDEBUG) {
        cerr << "compare::NUMBER OF TYPES OF ATOMIC SPECIES:   xstr1:  " << xstr1.num_each_type.size()
             << " 		xstr2:  " << xstr2.num_each_type.size() << endl;
        cerr << "compare::NUMBER OF TYPES OF ATOMIC SPECIES IS NOT THE SAME..." << endl;
      }
      return false;
    }
  }

  //Number of matches for xstr1 and xstr2 should equal the number of species in xstr1 and xstr2
  if(xstr1.num_each_type.size() == matches) {
    if(display == true) {
      if(LDEBUG) {
        cerr << "compare::NUMBER AND TYPE OF ATOMIC SPECIES ARE THE SAME...PROCEEDING..." << endl;
      }
    }
    return true;
  } else {
    if(display == true) {
      if(LDEBUG) {
        cerr << "compare::WARNING:: TYPE OF ATOMIC SPECIES OR NUMBER PER TYPE ARE NOT THE SAME..." << endl;
      }
    }
    return false;
  }
}
}

// ***************************************************************************
// Rescale Structure
// ***************************************************************************
namespace compare {
void rescaleStructure(xstructure& xstr1, xstructure& xstr2) {
  // If the scale factor is different, the two structures are rescaled to 1.00

  if(abs(xstr1.scale - xstr2.scale) > 0.001) {
    xstr1.ReScale(1.0);
    xstr2.ReScale(1.0);
    xstr1.FixLattices();
    xstr2.FixLattices();
  }
}
}

// ***************************************************************************
// Atomic Number Desnity
// ***************************************************************************
namespace compare {
void atomicNumberDensity(xstructure& xstr1, xstructure& xstr2) {
  // In order to compare structure with different volumes
  // we rescale the cell so that the volume divided by
  // the number of atoms is the same.

  double scale;
  scale = (xstr1.Volume() / xstr1.atoms.size()) / (xstr2.Volume() / xstr2.atoms.size());
  xstr2.InflateVolume(scale);
}
}

// ***************************************************************************
// Fake Atoms Name
// ***************************************************************************
namespace compare {
void fakeAtomsName(xstructure& xstr) {
  // Assign a fake letter to each atom type. In case of materials with more
  // than 26 species it is necessary to add more characters to this string

  string letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
  int iat = 0;
  //TEST    uint i,j;

  for (uint i = 0; i < xstr.num_each_type.size(); i++) {
    xstr.species.at(i) = letters[i];
    for (int j = 0; j < xstr.num_each_type.at(i); j++) {
      xstr.atoms.at(iat).name = letters[i];
      iat++;
    }
  }
}
}

// ***************************************************************************
// Print Structure Parameters
// ***************************************************************************
namespace compare {
void printParameters(xstructure& xstr, ostream& oss) {
  oss << "========================================================" << endl;
  oss << xstr.title << endl;
  oss << "Parameters:" << endl;
  oss << "a:     " << xstr.a << endl;
  oss << "b:     " << xstr.b << endl;
  oss << "c:     " << xstr.c << endl;
  oss << "alpha: " << xstr.alpha << endl;
  oss << "beta:  " << xstr.beta << endl;
  oss << "gamma: " << xstr.gamma << endl;
  oss << "volume:" << xstr.GetVolume() << endl;
}
}

// ***************************************************************************
// Least Frequent Atom
// ***************************************************************************
namespace compare {
string leastFrequentAtom(const xstructure& xstr) {
  // Least frequent atom can be exploited in case we want to reduce the number
  // of quadruplets to compute the rotational matrix by considering only the
  // LFA species. It differs from other leastfrequentatom2 function because it
  // finds the first LFA atom in xstructure 1. We do not need to find more the
  // one LFA species in xstructure 1 because we will find all the LFAs for
  // xstructure 2 and if the structures are a match we are guaranteed to find
  // a one-to-one correspondence from one of the LFAs in xstructure 2.

  int flag = 0, leastFrequentAtomCount = 0;
  string leastFrequentAtomType;

  for (uint i = 0; i < xstr.num_each_type.size(); i++) {
    if(flag == 0) {
      // Number of occurrences of the LFA
      leastFrequentAtomCount = xstr.num_each_type.at(i);
      // LFA Species
      leastFrequentAtomType = xstr.species.at(i);
      flag = 1;
    } else {
      if(leastFrequentAtomCount > xstr.num_each_type.at(i)) {
        leastFrequentAtomCount = xstr.num_each_type.at(i);
        leastFrequentAtomType = xstr.species.at(i);
      }
    }
  }
  return leastFrequentAtomType;
}
}

// ***************************************************************************
// Least Frequent Atom 2
// ***************************************************************************
namespace compare {
vector<string> leastFrequentAtom2(const xstructure& xstr) {
  // This least frequent atom function finds all possible least frequent atoms
  // for an xstructure and stores them in a vector. All of these LFAs are used
  // in the quadruplet search.
  // We may not need to search over multiple LFAs during the quadruplet search.
  // If a match is not found for one LFA, it won't be found for another since
  // we need to map all atoms in one structure to the other structure. We will
  // leave this implementation in for now, but may speed up the quadruplet
  // search if we consider only one LFA.

  int flag = 0, leastFrequentAtomCount = 0;
  vector<string> leastFrequentAtomType;

  for (uint i = 0; i < xstr.num_each_type.size(); i++) {
    if(flag == 0) {
      // Number of occurrences of the LFA
      leastFrequentAtomCount = xstr.num_each_type.at(i);
      // LFA Species
      leastFrequentAtomType.push_back(xstr.species.at(i));
      flag = 1;
    } else {
      if(leastFrequentAtomCount > xstr.num_each_type.at(i)) {
        leastFrequentAtomCount = xstr.num_each_type.at(i);
        leastFrequentAtomType.clear();
        leastFrequentAtomType.push_back(xstr.species.at(i));
      }
      if(leastFrequentAtomCount == xstr.num_each_type.at(i) && leastFrequentAtomType.at(0) != xstr.species.at(i)) {
        // Added the statement after '&&' (above); ensures no double counting from previous if statement
        leastFrequentAtomType.push_back(xstr.species.at(i));
      }
    }
  }
  return leastFrequentAtomType;
}
}

// ***************************************************************************
// Check Tolerances
// ***************************************************************************
namespace compare {
bool checkTolerance(xvector<double> d1, xvector<double> d2) {
  // When we look for 2 corresponding reference frames, we have to check that
  // the length of the 3 vectors and the angles between them are equal or within
  // a given tolerance (in this case 10% of the reference value has been set).

  double tol_length = 0.1, tol_angle = 0.1;

  if(abs(d1(1) - d2(1)) < tol_length * abs(d1(1)) &&
      abs(d1(2) - d2(2)) < tol_length * abs(d1(2)) &&
      abs(d1(3) - d2(3)) < tol_length * abs(d1(3)) &&
      abs(d1(4) - d2(4)) < tol_angle * abs(d1(4)) &&
      abs(d1(5) - d2(5)) < tol_angle * abs(d1(5)) &&
      abs(d1(6) - d2(6)) < tol_angle * abs(d1(6))) {
    return false;
  } else {
    return true;
  }
}
}

// ***************************************************************************
// Check ABC Tolerances
// ***************************************************************************
namespace compare {
bool checkABCTolerance(xvector<double> d1, xvector<double> d2) {
  // Similar to checkTolerance, but it only looks at the length of the lattice
  // vectors (for screening structures; makes faster)

  double tol_length = 0.1;

  for (uint i = 1; i < 4; i++) {
    for (uint j = 1; j < 4; j++) {
      if(j != i) {
        for (uint k = 1; k < 4; k++) {
          if(k != i && k != j) {
            if(abs(d1(i) - d2(1)) < tol_length * abs(d1(i)) &&
                abs(d1(j) - d2(2)) < tol_length * abs(d1(j)) &&
                abs(d1(k) - d2(3)) < tol_length * abs(d1(k))) {
              return false;
            }
          }
        }
      }
    }
  }
  return true;
}
}

// ***************************************************************************
// Check Angle Tolerances
// ***************************************************************************
namespace compare {
bool checkAngleTolerance(xvector<double> d1, xvector<double> d2) {
  // Similar to checkTolerance, but it only looks at the angles betweeen the
  // lattice vectors (for screening structures; makes faster)

  double tol_angle = 0.1;

  for (uint i = 4; i < 7; i++) {
    for (uint j = 4; j < 7; j++) {
      if(j != i) {
        for (uint k = 4; k < 7; k++) {
          if(k != i && k != j) {
            if(abs(d1(i) - d2(4)) < tol_angle * abs(d1(4)) &&
                abs(d1(j) - d2(5)) < tol_angle * abs(d1(5)) &&
                abs(d1(k) - d2(6)) < tol_angle * abs(d1(6))) {
              return false;
            }
          }
        }
      }
    }
  }
  return true;
}
}

// ***************************************************************************
// Find Matches
// ***************************************************************************
namespace compare {
void findMatch(const xstructure& xstr1, const xstructure& PROTO,
               vector<uint>& im1, vector<uint>& im2) {
  // In order to find the best matchings the routine computes
  // the difference between one atom and all the others,
  // building a matrix of differences of coordinates.
  // Then, it checks which atoms have the best matching
  // with another atom in the second structure.

  // A | 1    A1, A2     I can check which is the best matching
  // B | 2 -> B1, B2  -> for the atom 1 and 2 in the structure
  // C |      C1, C2     with A,B,C,D
  // D |      D1, D2

  uint j, k;
  int i1, i2;  //Indices corresponding atoms
  double min;

  vector<double> vdiffs;               //Difference btwn atoms coords
  vector<vector<double> > all_vdiffs;  //For all the atoms

  im1.clear();
  im2.clear();
  vdiffs.clear();
  all_vdiffs.clear();

  xmatrix<double> klattice = xstr1.lattice;

  for (j = 0; j < PROTO.atoms.size(); j++) {
    vdiffs.clear();
    for (k = 0; k < xstr1.atoms.size(); k++) {
      double dist = 1e9;
      // Need to find the min distance; thus check distance between neighboring cells to find true minimum.
      for (int a = -2; a <= 2; a++) {
        for (int b = -2; b <= 2; b++) {
          for (int c = -2; c <= 2; c++) {
            dist = aurostd::min(dist, modulus(PROTO.atoms.at(j).cpos - xstr1.atoms.at(k).cpos + a * klattice(1) + b * klattice(2) + c * klattice(3)));
          }
        }
      }
      vdiffs.push_back(dist);
    }
    all_vdiffs.push_back(vdiffs);
  }

  if(xstr1.atoms.size() <= all_vdiffs.size()) {
    for (k = 0; k < xstr1.atoms.size(); k++) {
      for (j = 0; j < all_vdiffs.size(); j++) {
        if(j == 0) {
          min = all_vdiffs.at(j).at(k);
          i1 = k;
          i2 = j;
        } else {
          if(all_vdiffs.at(j).at(k) < min) {
            min = all_vdiffs.at(j).at(k);
            i1 = k;
            i2 = j;
          }
          if(abs(all_vdiffs.at(j).at(k) - min) < 0.00001) {
            min = all_vdiffs.at(j).at(k);
            i1 = k;
            i2 = j;
          }
        }
      }
      im1.push_back(i1);
      im2.push_back(i2);
    }
  } else {
    for (k = 0; k < all_vdiffs.size(); k++) {
      for (j = 0; j < xstr1.atoms.size(); j++) {
        if(j == 0) {
          min = all_vdiffs.at(k).at(j);
          i1 = j;
          i2 = k;
        } else {
          if(all_vdiffs.at(k).at(j) < min) {
            min = all_vdiffs.at(k).at(j);
            i1 = j;
            i2 = k;
          }
          if(abs(all_vdiffs.at(k).at(j) - min) < 0.00001) {
            min = all_vdiffs.at(k).at(j);
            i1 = j;
            i2 = k;
          }
        }
      }
      im1.push_back(i1);
      im2.push_back(i2);
    }
  }
}
}

// ***************************************************************************
// clusterize
// ***************************************************************************
namespace compare {
void clusterize(const xstructure& xstr1, const vector<uint>& im1, vector<string>& TYPE1,
                vector<uint>& QTA1, vector<vector<uint> >& I1) {
  // This function builds clusters/vectors of atoms of the same type. It is necessary when
  // we want to check correspondences between atoms of the same type or between atoms
  //of a specific species.

  uint i, j;
  int flag;
  vector<uint> i1;

  i1.clear();

  for (i = 0; i < im1.size(); i++) {
    i1.clear();
    if(i == 0) {
      TYPE1.push_back(xstr1.atoms.at(im1.at(i)).name);
      QTA1.push_back(1);
      i1.push_back(im1.at(i));
      I1.push_back(i1);
    } else {
      flag = 0;
      for (j = 0; j < TYPE1.size(); j++) {
        if(xstr1.atoms.at(im1.at(i)).name == TYPE1.at(j)) {
          QTA1.at(j)++;
          i1 = I1.at(j);
          i1.push_back(im1.at(i));
          I1.at(j) = i1;
          flag = 1;
          break;
        }
      }
      if(flag == 0) {
        TYPE1.push_back(xstr1.atoms.at(im1.at(i)).name);
        QTA1.push_back(1);
        i1.push_back(im1.at(i));
        I1.push_back(i1);
      }
    }
  }
}
}

// ***************************************************************************
// Same Atom Type
// ***************************************************************************
namespace compare {
bool sameAtomType(const xstructure& xstr1, const xstructure& xstr2, const vector<uint>& im1,
                  const vector<uint>& im2, const int& type_match) {
  // Bool function; checks when 2 atoms matched are OK

  uint i, j, k, w, z;
  vector<string> TYPE1, TYPE2;
  vector<uint> QTA1, QTA2;
  vector<vector<uint> > I1, I2;

  vector<int> flag, checkType;

  clusterize(xstr1, im1, TYPE1, QTA1, I1);
  clusterize(xstr2, im2, TYPE2, QTA2, I2);

  checkType.clear();

  for (i = 0; i < TYPE1.size(); i++) {
    for (j = 0; j < TYPE2.size(); j++) {
      if(QTA1.at(i) == QTA2.at(j)) {
        flag.clear();
        for (w = 0; w < I1.at(i).size(); w++) {
          for (z = 0; z < I2.at(j).size(); z++) {
            for (k = 0; k < im1.size(); k++) {
              //Means we want the same atomic species to be matched up.
              if(type_match == 2) {
                if(I1.at(i).at(w) == im1.at(k) && I2.at(j).at(z) == im2.at(k) &&
                    xstr1.atoms.at(im1.at(k)).name == xstr2.atoms.at(im2.at(k)).name)
                  flag.push_back(1);
              } else {
                if(I1.at(i).at(w) == im1.at(k) && I2.at(j).at(z) == im2.at(k)) flag.push_back(1);
              }
            }
          }
        }
        if(flag.size() == QTA1.at(i))
          checkType.push_back(1);
      }
    }
  }
  if(checkType.size() == TYPE1.size())
    return true;
  else
    return false;
}
}

// ***************************************************************************
// Clean Match
// ***************************************************************************
namespace compare {
bool cleanMatch(const vector<uint>& im1) {
  // The result of the findMatch function is a pair of vectors containing the indices of the matched
  // atoms of the structures. This function allows to delate the matchings where the same
  // index appears more the one time.

  uint i, j;

  for (i = 0; i < im1.size(); i++) {
    for (j = 0; j < im1.size(); j++) {
      if(i != j)
        if(im1.at(i) == im1.at(j))
          return true;
    }
  }
  return false;
}
}

// ***************************************************************************
// Cell Diagonal
// ***************************************************************************
namespace compare {
void cellDiagonal(const xstructure& xstr, vector<double>& diag_sum,
                  vector<double>& diag_diff, const double& scale) {
  // The cell diagonals are represented by the shortest and longest diagonals
  // (in the case of cubic lattice the 2 diagonals are equal)
  // They are obtained as the vectorial sum and difference of the lattice
  // basis vectors.

  xvector<double> origin;

  diag_sum.push_back(abs(distance(origin, (xstr.lattice(1) + xstr.lattice(2)) * scale)));
  diag_sum.push_back(abs(distance(origin, (xstr.lattice(1) + xstr.lattice(3)) * scale)));
  diag_sum.push_back(abs(distance(origin, (xstr.lattice(2) + xstr.lattice(3)) * scale)));
  diag_diff.push_back(abs(distance(xstr.lattice(1), xstr.lattice(2)) * scale));
  diag_diff.push_back(abs(distance(xstr.lattice(1), xstr.lattice(3)) * scale));
  diag_diff.push_back(abs(distance(xstr.lattice(2), xstr.lattice(3)) * scale));
}
}

// ***************************************************************************
// Lattice Deviation
// ***************************************************************************
namespace compare {
double latticeDeviation(const vector<double>& diag_sum1, const vector<double>& diag_sum2,
                        const vector<double>& diag_diff1, const vector<double>& diag_diff2) {
  // Lattice deviation is computed as the deviation of the 2 diagonals of each face
  // normalized on the sum of the diagonals of the faces of the reference structure.
  // The images of the diagonals of the mapped structure must be rescaled such that
  // the volume of the image of its unit cell of equals the volume of the unit
  // cell of the reference one.

  uint i;
  double d;
  vector<double> dev;

  for (i = 0; i < 3; i++)
    dev.push_back((abs(diag_sum1.at(i) - diag_sum2.at(i)) + abs(diag_diff1.at(i) - diag_diff2.at(i))) / (diag_sum2.at(i) + diag_diff2.at(i)));

  d = 1;
  for (i = 0; i < dev.size(); i++)
    d = d * (1 - dev.at(i));
  d = 1 - d;

  return d;
}
}

// ***************************************************************************
// Shortest Distance from one atom
// ***************************************************************************
namespace compare {
double shortestDistance(const xstructure& xstr, const uint& k) {
  // Find the distance with the closest distance

  //double min;
  double min_dist = 1e9;
  xmatrix<double> klattice = xstr.lattice;  //NEW
  for (uint ii = 0; ii < xstr.atoms.size(); ii++) {
    if(ii != k) {
      for (int a = -2; a <= 2; a++) {
        for (int b = -2; b <= 2; b++) {
          for (int c = -2; c <= 2; c++) {
            min_dist = aurostd::min(min_dist, modulus(xstr.atoms.at(k).cpos - xstr.atoms.at(ii).cpos + a * klattice(1) + b * klattice(2) + c * klattice(3)));
          }
        }
      }
    }
  }
  return min_dist;
}
}

// ***************************************************************************
// Coordinates Deviation
// ***************************************************************************
namespace compare {
void coordinateDeviation(const xstructure& xstr1, const xstructure& xstr2,
                         const vector<uint>& indexMatch1, const vector<uint>& indexMatch2,
                         double& cd, double& fail_figure) {
  // To compute the coordinates deviation we look at each pair of atoms from
  // the reference and mapped structure

  uint j;
  double num = 0, den = 0, nfail = 0;
  double dd, nn1, nn2;  //dd=delta distance, nn=nearest neighbour
  int fail1, fail2;
  xmatrix<double> klattice = xstr1.lattice;
  for (j = 0; j < indexMatch1.size(); j++) {
    nn1 = shortestDistance(xstr1, indexMatch1.at(j));
    nn2 = shortestDistance(xstr2, indexMatch2.at(j));
    dd = 1e9;
    // Need to find the min distance; thus check distance between neighboring cells to find true minimum.
    for (int a = -2; a <= 2; a++) {
      for (int b = -2; b <= 2; b++) {
        for (int c = -2; c <= 2; c++) {
          dd = aurostd::min(dd, modulus(xstr1.atoms.at(indexMatch1.at(j)).cpos - xstr2.atoms.at(indexMatch2.at(j)).cpos + a * klattice(1) + b * klattice(2) + c * klattice(3)));
        }
      }
    }

    if(dd <= 0.5 * nn1) fail1 = 0;
    if(dd > 0.5 * nn1) fail1 = 1;
    if(dd <= 0.5 * nn2) fail2 = 0;
    if(dd > 0.5 * nn2) fail2 = 1;

    if(fail1 == 0) {
      num = num + dd;
      den = den + nn1;
    }
    if(fail2 == 0) {
      num = num + dd;
      den = den + nn2;
    }
    if(fail1 == 1) nfail++;
    if(fail2 == 1) nfail++;
  }

  if(den == 0)
    cd = 1;
  else
    cd = num / den;

  //Consider unmatched atoms
  int flag = 0;
  for (uint i = 0; i < xstr1.atoms.size(); i++) {
    flag = 0;
    for (uint k = 0; k < indexMatch1.size(); k++) {
      if(i == indexMatch1.at(k)) {
        flag = 1;
      }
    }
    if(flag == 0) {  //Meaning this atom does not have a match; increase the figure of failure
      nfail++;
    }
  }
  for (uint i = 0; i < xstr2.atoms.size(); i++) {
    flag = 0;
    for (uint k = 0; k < indexMatch2.size(); k++) {
      if(i == indexMatch2.at(k)) {
        flag = 1;
      }
    }
    if(flag == 0) {  //Meaning this atom does not have a match; increase the figure of failure
      nfail++;
    }
  }

  fail_figure = (nfail / (xstr1.atoms.size() + xstr2.atoms.size()));
}
}

// ***************************************************************************
// Compute Misfit
// ***************************************************************************
namespace compare {
double computeMisfit(const double& dev, const double& dis, const double& fail) {
  // Combines differences between all aspects of crystal structure into a
  // figure of misfit. (See Burzlaff)

  double mis;

  mis = 1 - ((1 - dev) * (1 - dis) * (1 - fail));
  return mis;
}
}

// ***************************************************************************
// Print Matching Between Atoms
// ***************************************************************************
namespace compare {
void printMatch(const vector<uint>& indexMatch1, const vector<uint>& indexMatch2,
                const xstructure& PROTO, const xstructure& xstr1, ostream& oss) {
  // With this function we print the atoms matched in the previous function
  // whose indices are stored in the indexMatch vector.
  // This allows us to call them directly.
  // Lastly, the translational term, specific of each mapping, is printed.

  uint i, j;

  oss << "              Reference                               Mapped" << endl;
  for (j = 0; j < indexMatch1.size(); j++) {
    oss << indexMatch1.at(j) << "-" << indexMatch2.at(j) << "    ";
    oss << xstr1.atoms.at(indexMatch1.at(j)).cpos << " " << xstr1.atoms.at(indexMatch1.at(j)).name;
    oss << "       ";
    oss << PROTO.atoms.at(indexMatch2.at(j)).cpos << " " << PROTO.atoms.at(indexMatch2.at(j)).name << endl;
  }

  int flag = 0;

  oss << "----------------------------------------------------" << endl;
  oss << "Missing Atoms in Reference structure:" << endl;
  for (j = 0; j < xstr1.atoms.size(); j++) {
    flag = 0;
    for (i = 0; i < indexMatch1.size(); i++) {
      if(j == indexMatch1.at(i))
        flag = 1;
    }
    if(flag == 0) {
      oss << "# " << j << "   " << xstr1.atoms.at(j).cpos << "   " << xstr1.atoms.at(j).name << endl;
    }
  }

  oss << "Missing Atoms in Mapped structure:" << endl;
  for (j = 0; j < PROTO.atoms.size(); j++) {
    flag = 0;
    for (i = 0; i < indexMatch2.size(); i++) {
      if(indexMatch2.at(i) == j) flag = 1;
    }
    if(flag == 0) oss << "# " << j << "   " << PROTO.atoms.at(j).cpos << "   " << PROTO.atoms.at(j).name << endl;
  }
}
}

// ***************************************************************************
// Bring Coordinate in the cell (Similar to xstructure.BringInCell())
// ***************************************************************************
namespace compare {
xvector<double> bringCoordinateInCell(xvector<double>& coord) {
  // This function brings a coordinate back in the unit cell.
  // It is not an xstructure attribute, so we cannot use xstructure.BringInCell().

  double tol = 1e-6;
  for (uint i = 1; i < 4; i++) {
    if(coord(i) < -tol) {
      coord(i) += 1.0;
    } else if(coord(i) >= 1.0 - tol) {
      coord(i) -= 1.0;
    }
  }
  return coord;
}
}

// ***************************************************************************
// Determine if LFA Quadruplet is Periodic
// ***************************************************************************
namespace compare {
bool quadrupletPeriodic(const xmatrix<double>& quad, const xstructure& lfa_supercell,
                        const int& i, const int& j, const int& k, const int& w) {
  // Once we have a possible quadruplet (lattice), we need to make sure that this
  // choice of the primitive cell preserves the periodicity o the lattice.
  // Therefore, we check that each of the quadruplet atoms maps onto another atom
  // in the supercell. Helpful analogy: Lattice periodicty vs crystal periodicity.
  // The quadruplets form the lattice and in this function we check for lattice
  // periodicity. The misfit criteria checks the crystal periodicity.

  vector<int> ind(4);
  ind[0] = i, ind[1] = j;
  ind[2] = k;
  ind[3] = w;
  vector<xvector<double> > latt_vecs(3);
  latt_vecs[0] = quad(1);
  latt_vecs[1] = quad(2);
  latt_vecs[2] = quad(3);
  xvector<double> tmp;

  for (uint b = 0; b < ind.size(); b++) {
    for (uint c = 0; c < latt_vecs.size(); c++) {
      tmp = lfa_supercell.atoms.at(ind[b]).cpos + latt_vecs[c];
      bool match_found = false;
      for (uint d = 0; d < lfa_supercell.atoms.size(); d++) {
        if(abs(lfa_supercell.atoms.at(d).cpos(1) - tmp(1)) < 0.01 &&
            abs(lfa_supercell.atoms.at(d).cpos(2) - tmp(2)) < 0.01 &&
            abs(lfa_supercell.atoms.at(d).cpos(3) - tmp(3)) < 0.01) {  // Less than tenth of Angstrom?
          match_found = true;
          break;
        }
      }
      if(match_found == false) {
        xvector<double> tmp_frac = C2F(lfa_supercell.lattice, tmp);
        tmp_frac = bringCoordinateInCell(tmp_frac);
        xstructure lfa_supercell_tmp = lfa_supercell;
        for (uint f = 0; f < lfa_supercell_tmp.atoms.size(); f++) {
          lfa_supercell_tmp.atoms.at(f).fpos = C2F(lfa_supercell_tmp.lattice, lfa_supercell_tmp.atoms.at(f).cpos);
          if(abs(lfa_supercell_tmp.atoms.at(f).fpos(1) - tmp_frac(1)) < 0.01 &&
              abs(lfa_supercell_tmp.atoms.at(f).fpos(2) - tmp_frac(2)) < 0.01 &&
              abs(lfa_supercell_tmp.atoms.at(f).fpos(3) - tmp_frac(3)) < 0.01) {
            match_found = true;
            break;
          }
        }
        if(match_found == false) {
          return false;
        }
      }
    }
  }
  return true;
}
}

// ***************************************************************************
// Thread Generation (For parallel processing of quadruplets)
// ***************************************************************************
namespace compare {
void threadGeneration(const int& num_proc, const xmatrix<double>& q1, xstructure& xstr2,
                      vector<xstructure>& vprotos, xstructure& xstr1, const int& type_match,
                      double& minMis, ostream& oss) {
  // This function creates the supercell of the second structure and begins
  // the quadruplets search within a supercell. Due to the costly nature of
  // this algorithm, the quadruplet search is parallelized. The splitting
  // of computation of done here

  bool LDEBUG = (false || XHOST.DEBUG);
  vector<string> LFA = leastFrequentAtom2(xstr2);
  string lfa;

  xstructure xstr = GetSuperCell(xstr2, 3, 0, 0, 0, 3, 0, 0, 0, 3);

  uint y;

  // Consider all LFAs
  for (y = 0; y < LFA.size(); y++) {
    lfa = LFA.at(y);
    oss << "===> LFA: " << lfa << endl;

    if(LDEBUG) {
      cerr << "===> LFA: " << lfa << endl;
    }

    // NEED TO SHIFT origin of xstr1 to one of the LFA (this was missing before and caused ICSD_102428.BCA, and CBA to not match, but they should
    for (uint i = 0; i < xstr1.atoms.size(); i++) {
      if(xstr1.atoms[i].name == lfa) {
        xstr1.ShifOriginToAtom(i);
        xstr1.BringInCell();
        break;
      }
    }

    // NEED TO SHIFT origin of xstr2 to one of the LFA
    for (uint i = 0; i < xstr2.atoms.size(); i++) {
      if(xstr2.atoms[i].name == lfa) {
        xstr2.ShifOriginToAtom(i);
        xstr2.BringInCell();
        break;
      }
    }

    // When checking the quadruplets/lattice, we only need to generate a
    // LFA supercell (i.e. take out the other atoms).  This greatly reduces
    // the time of computation (don't need to scan through unnecessary atoms)
    xstructure xstr_LFA_only = xstr2;
    uint num_atoms = xstr_LFA_only.atoms.size();
    for (uint i = 0; i < num_atoms; i++) {
      if(xstr_LFA_only.atoms.at(i).name != lfa) {
        xstr_LFA_only.RemoveAtom(i);
        num_atoms--;
        i--;
      }
    }
    xstructure xstr_LFA_supercell = GetSuperCell(xstr_LFA_only, 3, 0, 0, 0, 3, 0, 0, 0, 3);

    // Determines the number of LFAs in the supercell.
    int num_LFAs = -1;  //-1 as default value
    for (uint q = 0; q < xstr.num_each_type.size(); q++) {
      if(xstr.species.at(q) == LFA.at(y)) {
        num_LFAs = xstr.num_each_type.at(q);
        if(LDEBUG) { cerr << "compare:: "
                           << "Number of LFAs in supercell: " << xstr.species.at(q) << "= " << num_LFAs << endl; }
      }
    }

    // === THREAD PREPARATION FOR PARALLEL PROCESSING OF QUADRUPLET SEARCH === //

    // DECLARATION OF ATOMIC BOOL SECTION: Allows the threads to communicate.
    // This is useful for stopping the threads if the misfit falls below
    // the compatible misfit criterion (mis<0.1) in any of the threads.
    //[THREADS]      std::atomic_bool misfit_in_threshold_found (false);

    bool misfit_in_threshold_found = false;

    //DYNAMIC THREAD GENERATION
    vector<uint> min_index_i;
    vector<uint> max_index_i;
    vector<uint> min_index_j;
    vector<uint> max_index_j;
    vector<uint> min_index_k;
    vector<uint> max_index_k;
    vector<uint> min_index_w;
    vector<uint> max_index_w;
    vector<xstructure> xstr1_for_thread;
    vector<double> possible_minMis;
    vector<vector<xstructure> > vvprotos;
    //[THREADS]      vector<std::thread> threads;
    //[THREADS]      int num_leftover=0;

    // First, we determine the possible number of quadruplet combinations which
    // can be formed from the number of LFAs.  To do this, we use the typical
    // combination formula: C(n,r) = n!/r!(n-r)! . To simplify this calcuation,
    // we know that r will always be 4 since we are finding quadruplets (4 atoms
    // forming a lattice).  This means that that the quantity (n-r) will always
    // be n-4.  Therefore, expanding out the factorials and canceling like terms
    // in the numerator and denominator reduces the combination equation to :
    // C(n,r) = (n)(n-1)(n-2)(n-3)/4!.  This is done below.

    // COMBINATION CALCULATION
    unsigned long long int combos = 0;
    unsigned long long int numerator = 1;  // Set to 1, since we will use multiplication
    int denominator = 1;

    for (uint c = 0; c < 4; c++) {
      numerator *= (num_LFAs - c);
      denominator *= (4 - c);
    }
    combos = numerator / denominator;
    // END OF COMBINATION CALCULATION

    //Determine the number of quadruplets per number of processors.
    unsigned long long int num_combos_per_thread = combos / num_proc;
    unsigned long long int left_over = combos % num_proc;  //For uneven division, tack on the last quadruplets to the last thread/processor.
    if(LDEBUG) {
      cerr << "compare:: "
           << "Number of combinations to search per thread= " << num_combos_per_thread;
      cerr << "; Leftovers added to first thread= " << left_over << endl;
    }

    // ===  Splitting Quadruplets Among Threads: Straightforward method === //

    unsigned long long int total_combo = 0;
    unsigned long long int number = 0;
    int thread_number = 0;
    vector<vector<uint> > min_indices;
    vector<vector<uint> > max_indices;
    max_indices.resize(num_proc);
    min_indices.resize(num_proc);
    min_indices.at(thread_number).push_back(0);
    min_indices.at(thread_number).push_back(1);
    min_indices.at(thread_number).push_back(2);
    min_indices.at(thread_number).push_back(3);
    if(num_proc != 1) {
      for (uint i = 0; i < xstr_LFA_supercell.atoms.size(); i++) {
        for (uint j = i + 1; j < xstr_LFA_supercell.atoms.size(); j++) {
          for (uint k = j + 1; k < xstr_LFA_supercell.atoms.size(); k++) {
            for (uint w = k + 1; w < xstr_LFA_supercell.atoms.size(); w++) {
              number++;
              if(number == num_combos_per_thread) {
                total_combo += number;
                number = 0;
                max_indices.at(thread_number).push_back(i);
                max_indices.at(thread_number).push_back(j);
                max_indices.at(thread_number).push_back(k);
                max_indices.at(thread_number).push_back(w);
                thread_number++;
                if(thread_number == num_proc - 1) {
                  num_combos_per_thread += left_over;
                }
                if(thread_number < num_proc) {
                  if(w != xstr_LFA_supercell.atoms.size() - 1) {
                    // Store the starting indices for the next thread
                    min_indices.at(thread_number).push_back(i);
                    min_indices.at(thread_number).push_back(j);
                    min_indices.at(thread_number).push_back(k);
                    min_indices.at(thread_number).push_back(w + 1);
                  } else if(w == xstr_LFA_supercell.atoms.size() - 1) {
                    if(k != xstr_LFA_supercell.atoms.size() - 2) {
                      // Store the starting indices for the next thread
                      min_indices.at(thread_number).push_back(i);
                      min_indices.at(thread_number).push_back(j);
                      min_indices.at(thread_number).push_back(k + 1);
                      min_indices.at(thread_number).push_back(k + 2);
                    } else if(k == xstr_LFA_supercell.atoms.size() - 2) {
                      if(j != xstr_LFA_supercell.atoms.size() - 3) {
                        min_indices.at(thread_number).push_back(i);
                        min_indices.at(thread_number).push_back(j + 1);
                        min_indices.at(thread_number).push_back(j + 2);
                        min_indices.at(thread_number).push_back(j + 3);
                      } else if(j == xstr_LFA_supercell.atoms.size() - 3) {
                        min_indices.at(thread_number).push_back(i + 1);
                        min_indices.at(thread_number).push_back(i + 2);
                        min_indices.at(thread_number).push_back(i + 3);
                        min_indices.at(thread_number).push_back(i + 4);
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    } else {
      max_indices.at(thread_number).push_back(num_LFAs - 4);
      max_indices.at(thread_number).push_back(num_LFAs - 3);
      max_indices.at(thread_number).push_back(num_LFAs - 2);
      max_indices.at(thread_number).push_back(num_LFAs - 1);
    }
    //DEBUGGING
    //cerr << "total= " << total_combo << endl;
    //cerr << "min_indices.size()= " << min_indices.size()
    //     << "max_indices.size()= " << max_indices.size() <<endl;
    //for(uint q=0; q<min_indices.size(); q++){
    //  for(uint t=0; t<min_indices.at(q).size();t++){
    //    cerr << "min_indices.at(" << q<< ").at(" << t << ")= " i
    //    << min_indices.at(q).at(t) << "; max_indices.at(" << q<< ").at(" << t << ")= "
    //    << max_indices.at(q).at(t) <<  endl;
    //  }
    //}
    //DEBUGGING

    for (int n = 0; n < num_proc; n++) {
      vector<xstructure> vprotos_tmp;
      vvprotos.push_back(vprotos_tmp);
      xstr1_for_thread.push_back(xstr1);
      possible_minMis.push_back(1.0);
    }

    //[THREADS]      // Run threads
    //[THREADS]      for(uint n=0; n<num_proc; n++){
    //[THREADS]	threads.push_back(std::thread(quadrupletSearch,std::ref(misfit_in_threshold_found),lfa,min_indices.at(n).at(0),
    //[THREADS]			  max_indices.at(n).at(0),min_indices.at(n).at(1),max_indices.at(n).at(1),min_indices.at(n).at(2),
    //[THREADS]			  max_indices.at(n).at(2),min_indices.at(n).at(3),max_indices.at(n).at(3),q1,xstr,xstr_LFA_supercell,
    //[THREADS]			  std::ref(vvprotos.at(n)),std::ref(xstr1_for_thread.at(n)),xstr2,type_match,std::ref(possible_minMis.at(n)),
    //[THREADS]			  std::ref(oss)));
    //[THREADS]      }

    //[THREADS]      // Wait until all threads are finished or stopped
    //[THREADS]      for(uint t=0;t<threads.size();t++){
    //[THREADS]	threads.at(t).join();
    //[THREADS]      }

    //Non-threaded
    uint n = 0;
    quadrupletSearch(misfit_in_threshold_found, lfa, min_indices.at(n).at(0), max_indices.at(n).at(0),
                     min_indices.at(n).at(1), max_indices.at(n).at(1), min_indices.at(n).at(2),
                     max_indices.at(n).at(2), min_indices.at(n).at(3), max_indices.at(n).at(3), q1, xstr,
                     xstr_LFA_supercell, vvprotos.at(n), xstr1_for_thread.at(n), xstr2, type_match, possible_minMis.at(n));

    //Non-threaded end
    for (uint p = 0; p < possible_minMis.size(); p++) {
      if(p == 0 && y == 0) {
        minMis = possible_minMis.at(p);
        xstr1 = xstr1_for_thread.at(p);
        vprotos = vvprotos.at(p);
      } else {
        if(possible_minMis.at(p) <= minMis) {
          minMis = possible_minMis.at(p);
          xstr1 = xstr1_for_thread.at(p);
          vprotos = vvprotos.at(p);
        }
      }
    }
    if(minMis <= 0.1) break;
  }
}
}

// ***************************************************************************
// Quadruplet Search
// ***************************************************************************
namespace compare {
//[THREADS]  void quadrupletSearch(std::atomic_bool &misfit_in_threshold_found, const string& lfa,
//[THREADS]			const int& min_index_i, const int& max_index_i,const int& min_index_j,
//[THREADS]			const int& max_index_j,const int& min_index_k, const int& max_index_k,
//[THREADS]			const int& min_index_w,const int& max_index_w, const xmatrix<double>& q1,
//[THREADS]			const xstructure& xstr, const xstructure& xstr_LFA_supercell,
//[THREADS]			vector<xstructure>& vprotos, xstructure& xstr1, const xstructure& xstr2,
//[THREADS]			const int& type_match, double& possible_minMis, ostream& oss){ // 1 bracket
void quadrupletSearch(bool& misfit_in_threshold_found, const string& lfa, const uint& min_index_i,
                      const uint& max_index_i, const uint& min_index_j, const uint& max_index_j, const uint& min_index_k,
                      const uint& max_index_k, const uint& min_index_w, const uint& max_index_w, const xmatrix<double>& q1,
                      const xstructure& xstr, const xstructure& xstr_LFA_supercell, vector<xstructure>& vprotos,
                      xstructure& xstr1, const xstructure& xstr2, const uint& type_match, double& possible_minMis) {
  // This function scans through the possible quadruplets (sets of 4 LFA atoms) i
  // to find a lattice which is commensurate with the reference structure (xstr1).
  // This function is parallelized since it is the time-limiting function.

  //bool LDEBUG=(false || XHOST.DEBUG);
  uint i, j, k, w, z, iat;
  xmatrix<double> q2(3, 3);
  double tol_vol = 0.1;
  double mis = 1;
  xstructure proto;
  int flag = 0;

  // Calculate xstr1 properties here; saves computation time
  // (which is valuable when there are many combinations)

  // Calculate cell diagonals
  vector<double> D1, F1;
  cellDiagonal(xstr1, D1, F1, 1);

  // Calculate determinant of q1 lattice and tolerance
  double abs_det_q1 = abs(det(q1));
  double det_tol = tol_vol * abs_det_q1;

  xvector<double> abc_angles_q1 = Getabc_angles(q1, DEGREES);

  xstr1.lattice = GetClat(xstr1.a, xstr1.b, xstr1.c, xstr1.alpha, xstr1.beta, xstr1.gamma);
  for (iat = 0; iat < xstr1.atoms.size(); iat++)
    xstr1.atoms.at(iat).cpos = F2C(xstr1.lattice, xstr1.atoms.at(iat).fpos);

  // Determine how much to shift the index to compare the original supercell and the LFA supercell
  int atom_index_shift = 0;
  for (uint q = 0; q < xstr.num_each_type.size(); q++) {
    if(xstr.species.at(q) == lfa) {
      for (uint p = 0; p < q; p++) {
        atom_index_shift += xstr.num_each_type.at(p);
      }
    }
  }

  bool j_continue = false;
  bool k_continue = false;
  bool w_continue = false;
  unsigned long long int counting = 0;
  //The following nested loops determine all the possible quadruplets formed by four
  // of the LFA atoms (restricted search to avoid duplicates).
  for (i = min_index_i; i <= max_index_i; i++) {                   // 2 bracket
    if(xstr_LFA_supercell.atoms.at(i).name == lfa) {              // 3 bracket
      for (j = i + 1; j < xstr_LFA_supercell.atoms.size(); j++) {  // 5 bracket
        if((i == min_index_i && j >= min_index_j)) {
          j_continue = true;
        }
        if(i == max_index_i && j > max_index_j) {
          j_continue = false;
        }
        if(j_continue == true) {                                      // 7 bracket
          for (k = j + 1; k < xstr_LFA_supercell.atoms.size(); k++) {  // 8 bracket
            if((i == min_index_i && j >= min_index_j && k >= min_index_k)) {
              k_continue = true;
            }
            if(i == max_index_i && j == max_index_j && k > max_index_k) {
              k_continue = false;
            }
            if(k_continue) {                                              //10 bracket
              for (w = k + 1; w < xstr_LFA_supercell.atoms.size(); w++) {  //11 bracket  //DX added for test here
                if(misfit_in_threshold_found == false) {                  //12 bracket
                  //DX added atomic boolean here since this is the deepest nested place (checked frequently)
                  if((i == min_index_i && j == min_index_j && k == min_index_k && w >= min_index_w)) {
                    w_continue = true;
                  }
                  if(i == max_index_i && j == max_index_j && k == max_index_k && w > max_index_w) {
                    w_continue = false;
                  }
                  if(w_continue == true) {  //14 bracket
                    counting++;
                    for (z = 1; z < 4; z++) {
                      q2(1, z) = xstr_LFA_supercell.atoms.at(j).cpos(z) - xstr_LFA_supercell.atoms.at(i).cpos(z);
                      q2(2, z) = xstr_LFA_supercell.atoms.at(k).cpos(z) - xstr_LFA_supercell.atoms.at(i).cpos(z);
                      q2(3, z) = xstr_LFA_supercell.atoms.at(w).cpos(z) - xstr_LFA_supercell.atoms.at(i).cpos(z);
                    }
                    // Determines if quadruplet is consistent with volume of xstructure1 lattice
                    // Since we are working with the same quadruplets the volume formed by these
                    // 4 atoms will not change. This saves time so we do not find all the orientations
                    // for quadruplets whose volumes are inconsistent with the criterion.
                    if(abs(abs_det_q1 - abs(det(q2))) < det_tol) {  //15 bracket
                      int at_origin = 1;
                      while (at_origin < 5) {
                        if(at_origin == 1) {
                          //kept the same as the test above
                        }
                        if(at_origin == 2) {
                          for (z = 1; z < 4; z++) {
                            q2(1, z) = xstr_LFA_supercell.atoms.at(i).cpos(z) - xstr_LFA_supercell.atoms.at(j).cpos(z);
                            q2(2, z) = xstr_LFA_supercell.atoms.at(k).cpos(z) - xstr_LFA_supercell.atoms.at(j).cpos(z);
                            q2(3, z) = xstr_LFA_supercell.atoms.at(w).cpos(z) - xstr_LFA_supercell.atoms.at(j).cpos(z);
                          }
                        }
                        if(at_origin == 3) {
                          for (z = 1; z < 4; z++) {
                            q2(1, z) = xstr_LFA_supercell.atoms.at(i).cpos(z) - xstr_LFA_supercell.atoms.at(k).cpos(z);
                            q2(2, z) = xstr_LFA_supercell.atoms.at(j).cpos(z) - xstr_LFA_supercell.atoms.at(k).cpos(z);
                            q2(3, z) = xstr_LFA_supercell.atoms.at(w).cpos(z) - xstr_LFA_supercell.atoms.at(k).cpos(z);
                          }
                        }
                        if(at_origin == 4) {
                          for (z = 1; z < 4; z++) {
                            q2(1, z) = xstr_LFA_supercell.atoms.at(i).cpos(z) - xstr_LFA_supercell.atoms.at(w).cpos(z);
                            q2(2, z) = xstr_LFA_supercell.atoms.at(j).cpos(z) - xstr_LFA_supercell.atoms.at(w).cpos(z);
                            q2(3, z) = xstr_LFA_supercell.atoms.at(k).cpos(z) - xstr_LFA_supercell.atoms.at(w).cpos(z);
                          }
                        }
                        xvector<double> abc_angles_q2 = Getabc_angles(q2, DEGREES);
                        if(checkABCTolerance(abc_angles_q1, abc_angles_q2) == false) {
                          xmatrix<double> tmp_quad(3, 3);
                          for (uint l = 1; l < 4; l++) {
                            for (uint m = 1; m < 4; m++) {
                              if(m != l) {
                                for (uint n = 1; n < 4; n++) {
                                  if(n != m && n != l) {
                                    for (z = 1; z < 4; z++) {
                                      tmp_quad(l, z) = q2(1, z);
                                      tmp_quad(m, z) = q2(2, z);
                                      tmp_quad(n, z) = q2(3, z);
                                    }
                                    abc_angles_q2 = Getabc_angles(tmp_quad, DEGREES);
                                    // Check abc and angles
                                    if(checkTolerance(abc_angles_q1, abc_angles_q2) == false) {
                                      // Make sure the quadruplet is periodic
                                      if(quadrupletPeriodic(tmp_quad, xstr_LFA_supercell, i, j, k, w) == true) {
                                        proto = xstr;
                                        proto.lattice = tmp_quad;
                                        if(at_origin == 1) {
                                          proto.ShifOriginToAtom(i + atom_index_shift);
                                        }
                                        if(at_origin == 2) {
                                          proto.ShifOriginToAtom(j + atom_index_shift);
                                        }
                                        if(at_origin == 3) {
                                          proto.ShifOriginToAtom(k + atom_index_shift);
                                        }
                                        if(at_origin == 4) {
                                          proto.ShifOriginToAtom(w + atom_index_shift);
                                        }
                                        for (iat = 0; iat < proto.atoms.size(); iat++) {
                                          proto.atoms.at(iat).fpos = C2F(proto.lattice, proto.atoms.at(iat).cpos);
                                        }
                                        proto.lattice = GetClat(abc_angles_q2[1], abc_angles_q2[2], abc_angles_q2[3],
                                                                abc_angles_q2[4], abc_angles_q2[5], abc_angles_q2[6]);
                                        for (iat = 0; iat < proto.atoms.size(); iat++) {
                                          proto.atoms.at(iat).cpos = F2C(proto.lattice, proto.atoms.at(iat).fpos);
                                        }
                                        uint N = proto.atoms.size();
                                        for (iat = 0; iat < N; iat++) {
                                          if(proto.atoms.at(iat).fpos(1) < -0.000001 || proto.atoms.at(iat).fpos(1) > 0.999999 ||
                                              proto.atoms.at(iat).fpos(2) < -0.000001 || proto.atoms.at(iat).fpos(2) > 0.999999 ||
                                              proto.atoms.at(iat).fpos(3) < -0.000001 || proto.atoms.at(iat).fpos(3) > 0.999999) {
                                            proto.RemoveAtom(iat);
                                            N--;
                                            iat--;
                                          }
                                        }
                                        // Ensure the new representation (proto) has the same number of species as xstr1 or xstr2
                                        if(sameSpecies(proto, xstr2, false) || sameSpecies(proto, xstr1, false)) {
                                          for (iat = 0; iat < proto.atoms.size(); iat++) {
                                            if(proto.atoms.at(iat).name == lfa) {
                                              proto.ShifOriginToAtom(iat);
                                              proto.BringInCell();
                                              vector<uint> im1, im2;
                                              findMatch(xstr1, proto, im1, im2);
                                              if(cleanMatch(im2) == false && cleanMatch(im1) == false) {
                                                if(sameAtomType(xstr1, proto, im1, im2, type_match) == true) {
                                                  double scale = xstr1.Volume() / proto.Volume();
                                                  scale = pow(scale, 0.3333);
                                                  vector<double> D2, F2;
                                                  cellDiagonal(proto, D2, F2, scale);
                                                  double latt_dev = latticeDeviation(D1, D2, F1, F2);
                                                  double cd, f;
                                                  coordinateDeviation(xstr1, proto, im1, im2, cd, f);
                                                  mis = computeMisfit(latt_dev, cd, f);
                                                  if(flag == 0) {
                                                    flag = 1;
                                                    vprotos.push_back(proto);
                                                    possible_minMis = mis;
                                                  } else {
                                                    if(mis < possible_minMis) {
                                                      vprotos.clear();
                                                      possible_minMis = mis;
                                                      vprotos.push_back(proto);  //to here
                                                    }
                                                  }
                                                  if(mis < 0.1) {
                                                    //DEBUGGING
                                                    //cerr <<"Winning combo: "<<i<<","<<j<<","<<k<<","<<w<<endl;
                                                    //cerr << "proto.lattice: " << proto.lattice << endl;
                                                    //cerr << "lattice(1): " << modulus(proto.lattice(1)) << endl;
                                                    //cerr << "lattice(2): " << modulus(proto.lattice(2)) << endl;
                                                    //cerr << "lattice(3): " << modulus(proto.lattice(3)) << endl;
                                                  }
                                                  if(mis < 0.1) break;
                                                }
                                              }
                                            }
                                          }
                                        }  // end of if protos.size()...
                                      }
                                    }
                                    if(mis < 0.1) break;
                                  }
                                  if(mis < 0.1) break;
                                }
                                if(mis < 0.1) break;
                              }
                              if(mis < 0.1) break;
                            }
                            if(mis < 0.1) break;
                          }
                          if(mis < 0.1) break;  //DX
                        }
                        at_origin++;
                      }
                      if(mis < 0.1) break;
                    }
                    if(mis < 0.1) break;  //15 bracket
                  }
                  if(mis < 0.1) break;  //14 bracket
                }                        //12 bracket
              }
              if(mis < 0.1 || misfit_in_threshold_found == true) break;  // 11 bracket //DX changed from 0.01 to 0.1 (SEE BURZLAFF PAPER)
            }                                                             // 10 bracket
          }
          if(mis < 0.1 || misfit_in_threshold_found == true) break;  //8 bracket   //DX changed from 0.01 to 0.1 (SEE BURZLAFF PAPER)
        }                                                             // 7 bracket
      }
      if(mis < 0.1 || misfit_in_threshold_found == true) break;  // 5 bracket //DX changed from 0.01 to 0.1 (SEE BURZLAFF PAPER)
    }                                                             // 3 bracket
  }                                                               //if(mis<0.1)break; // 2 bracket

  //FOR DEBUGGING
  //      cerr << "compare::THREAD DONE. min_index_i: " << min_index_i << " max_index_i: " << max_index_i
  //	   << "; min_index_j: " << min_index_j << " max_index_j: " << max_index_j
  //	   << "; min_index_k: " << min_index_k << " max_index_k: " << max_index_k
  //	   << " min_index_w: " << min_index_w << " max_index_w: " << max_index_w << endl;
  //FOR DEBUGGING cerr << "compare:: " << "Number of combinations explored in thread =" << counting << endl;
  if(mis < 0.1) {
    misfit_in_threshold_found = true;
    // FOR DEBUGGING cerr << "compare:: " << "THE OTHER THREAD FOUND A MATCH BEFORE THIS ONE " << endl;
  }
}  //1 bracket
}

//---------------------------------------------------------------

// ***************************************************************************
//                                 END - FINE
// 			AFLOW Compare Structure - Functions
//		David Hicks (d.hicks@duke.edu) and Carlo de Santo
// ***************************************************************************
